library(compiler)

# ---- fonction preparation base de donnee pour modélisation ----
bdd2<-function(dat,V, metalsol,metalgrain,seuils=NULL){
  x=lc(bd=dat,
       c(metalsol,
         V
       )
  )
  
  x$don$Variete<-dat$Variete
  x$don$qtl<-dat$qtl
  x$don$Code=dat$Code

  
  if (!(is.null(seuils))) {
    coded=do.call('cbind',lapply(seuils,function(y){
      a=as.data.frame(rt(dat[,metalgrain],y))
      names(a)=paste0(
        substr(metalgrain,1,2),gsub(".","",as.character(y),fixed=T)
      )
      a
    })
    )

  }
  
  out=data.frame(dat[,
                     c(metalsol,
                       V,
                       metalgrain
                     )],
                 x$don)
  return(list(
    bd=if (!(is.null(seuils))) {data.frame(out,coded)} else {out},
    median_log10=x$median_log10))
}


# ---- fonction codage seuil ----

rt<-function(v, # vecteur à coder
             s # valeur du seuil (numeric)
             )
  {
    coded=rep(0,length(v))
    coded[which(v>s)]=1

    print(table(coded))
    return(coded)
}







# ---- fonction Log10 et centrage par la médiane ----
lc<-function(bd, # base de données
             var # variables à transformer et centrer
             ){
  bd.lc=data.frame(
    apply(bd[,c(var)],
          2,
          function(x){
            a=log10(x)
            m=median(a,na.rm=T)
            a-m
          })
  )
  
  # ajout du suffixe .lc aux variables transformées
  names(bd.lc)=paste(names(bd.lc),"lc",sep=".")
  
return(list(don=  data.frame(bd.lc),
            median_log10=apply(bd[,c(var)],2,
                               function(x){
                                 a=log10(x)
                                 median(a,na.rm=T)
                                 
                               })))

}








# ---- fonction pour vérifier un Kfoldesign ----
check.doe<-function(doe,y.coded){
  ck=as.data.frame(
    do.call('rbind',
          lapply(doe,function(x){
    table(y.coded[x$test])})
  )
  )
  summary(ck)

}
# ---- fonction générant toutes les combinaisons de prédicteurs ----
library(combinat)


Xi<-function(P # vecteur des prédicteurs
             ){
  x=do.call('rbind',
            lapply(1:length(P),function(i){
              a=combn(P,i,simplify = F)
              do.call('rbind',lapply(a,function(y){
                do.call('paste',c(y,sep=" + "))
              })
              )
            })
  )
x
}




# ---- fonction générant les formules de prédicteurs ----
build_mod<-function(ET_sol,add_var){
  
  # variables ne pouvant être ensemble
  
  b=c("Arg.lc",
      "Arg.lc + Limf.lc",
      "Arg.lc + Lim.lc")
  
  
  
  # combinaisons des autres facteurs hors Cd et pH qui seront forcés
  add_var.lc<-paste0(add_var,".lc")
  d=Xi(as.list(add_var.lc))
  
  
  # toutes les combinaisons de b et d
  e=expand.grid(d,b)
  f=paste(e[,1],e[,2], sep=" + ")
  g=c(b,d,f)
  
  # forçage du ET_sol et pH
  ET_sol.lc=paste0(ET_sol,".lc")
  Xiform0=c(ET_sol.lc,
            paste(ET_sol.lc, "Heau.lc",sep=" + "),

            paste(ET_sol.lc, "Heau.lc",g, sep=" + "))
  
  rm(b,d,e,f,g)
  
  
  
  Xiform.lc=Xiform0
  Xiform<-gsub(".lc","",Xiform.lc,fixed=T)
  
  
  return(list(Xiform.lc=Xiform.lc,
              Xiform=Xiform))
}
# ---- fonction pour fitter modèle linéaire mixte ----
fit.lme <- function(bd, # base de données
                    y, # variable dépendente (caractères)
                    P, # vecteurs des formules des prédicteurs
                    rand # formule de l'effet aléatoire (caractère)
                    )
                    
                    {
                      library(nlme)
                      library(doParallel)
                      
                      
                      a=mclapply(P, function(x) {
                        mod = as.formula(paste0(y, "~", x)) # formule du modèle
                        rdm = as.formula(paste0("~", rand)) # formule des effets aléatoires


                        MEM=do.call("lme",args=list(
                          mod,
                          random=rdm,
                          data=bd,
                          control = list(
                            opt = "optim",
                            maxIter = 200,
                            returnObject = T
                          ),
                          na.action = "na.exclude",
                          weights = varPower(form =  ~ fitted(.))
                        ))
                        
                        
                        MEM
                      },
                      mc.cores = detectCores() - 1)
                      names(a) = P
                      a
                    }
                    


# ---- fonction pour tester modèle linéaire mixte sur  jeu de données de calibration ----
fittest_all.lme <- function(bd, # base de données
                    y, # variable dépendente (caractères)
                    y.class, # variable dépendente codée
                    P, # vecteurs des formules des prédicteurs
                    s, # seuil testé numeric
                    rand # formule de l'effet aléatoire (caractère)
)
  
{
  library(nlme)
  library(doParallel)
  
  
  a=mclapply(P, function(x) {
    mod = as.formula(paste0(y, "~", x)) # formule du modèle
    rdm = as.formula(paste0("~", rand)) # formule des effets aléatoires
    
    
    MEM=do.call("lme",args=list(
      mod,
      random=rdm,
      data=bd,
      control = list(
        opt = "optim",
        maxIter = 200,
        returnObject = F
      ),
      na.action = "na.exclude",
      weights = varPower(form =  ~ fitted(.))
    ))
    
    if (exists("MEM")) {
      prednat=10^predict(MEM)
      mem.perf=perf(rt(v=prednat,s=s),bd[,y.class])
      fit=MEM
    } else {
      
      mem.perf=NA
      fit=NA
    }

    
    return(list(fit=fit,
                perf=mem.perf))
  },
  mc.cores = detectCores() - 1)
  names(a) = P
  a
}



# ---- fonction pour tester modèle multilinéaire  sur  jeu de données de calibration ----
fittest_all.lm <- function(bd, # base de données
                            y, # variable dépendente (caractères)
                            y.class, # variable dépendente codée
                            P, # vecteurs des formules des prédicteurs
                            s, # seuil testé numeric
                            categ # variable catégorielle (caractère)
)
  
{

  library(doParallel)
  
  
  a=mclapply(P, function(x) {
    mod = as.formula(paste0(y, "~", x ," + ", categ)) # formule du modèle
 
    
    
    MEM=lm(mod,data=bd,na.action = "na.exclude")

    
    if (exists("MEM")) {
      prednat=10^predict(MEM)
      mem.perf=perf(rt(v=prednat,s=s),bd[,y.class])
      fit=MEM
    } else {
      
      mem.perf=NA
      fit=NA
    }
    
    
    return(list(fit=fit,
                perf=mem.perf))
  },
  mc.cores = detectCores() - 1)
  names(a) = P
  a
}

# ---- fonction pour fitter modèle multilinéaire  ----
fit.lm <- function(bd, # base de données
                           y, # variable dépendente (caractères)
                           y.class, # variable dépendente codée
                           P, # une formules de prédicteurs
                           categ # variable catégorielle (caractère)
)
  
{
  

    mod = as.formula(paste0("log10(",y, ")~", P ," + ", categ)) # formule du modèle
    
    
    
    ML=lm(mod,data=bd,na.action = "na.exclude")
    ML


}

# ---- fonction pour fitter modèle logistic ----
fit.lg <- function(bd, # base de données
                    y, # variable dépendente (caractères)
                    P, # vecteurs des formules des prédicteurs
                    rand # formule de l'effet aléatoire (caractère)
)
  
{
  library(lme4)
  library(doParallel)
  
  
  a=mclapply(P, function(x) {
    tryCatch(
      {
    mod = as.formula(paste0(y, "~", x, "+(", rand,")")) # formule du modèle
    
    
    MLG = glmer(
      mod,
      data = bd,
      na.action = "na.exclude",
      family = binomial(link = "logit")
    )
    MLG
      },
    error=function(cond) {
      NA
    }
    )
  },
  mc.cores = detectCores() - 1)
  names(a) = P
  
  # enleve les non convergence qui son de class logique (na)
  a[!(sapply(a, is.logical))]
  
  
}

# ---- fonction pour fitter modèle logistic effets fixes ----
fit.lg.f <- function(bd, # base de données
                   y, # variable dépendente (caractères)
                   P, # UNE formule de prédicteurs
                   categ # formule de l'effet categorie (caractère)
)
  
{
  
 
    tryCatch(
      {

        
        
        MLG = glm(
          formula(paste0(y, "~", P, "+", categ)), # formule du modèle
          data = bd,
          na.action = "na.exclude",
          family = binomial(link = "logit")
        )
        
        MLG
      },
      error=function(cond) {
        NA
      }
    )


  
}



# ---- fonction pour tester modèle logistic sur  jeu de données de calibration ----
fittest_all.lg <- function(bd, # base de données
                   y.class, # variable dépendente (caractères)
                   P, # vecteurs des formules des prédicteurs
                   rand # formule de l'effet aléatoire (caractère)
)
  
{
  library(lme4)
  library(doParallel)
  
  
  a=mclapply(P, function(x) {
    tryCatch(
      {
        mod = as.formula(paste0(y.class, "~", x, "+(", rand,")")) # formule du modèle
        
        
        MLG = glmer(
          mod,
          data = bd,
          na.action = "na.exclude",
          family = binomial(link = "logit")
        )
        
        pred = predict(MLG, type = "response")
        bst=bestcutoff(pred,bd[,y.class])

        ctf.Y=bst$perf.Y$cutoff
        ctf.MCC=bst$perf.MCC$cutoff
        

        lg.test.Y=perf(rt(v=pred,s=ctf.Y),bd[,y.class])

        lg.test.MCC=perf(rt(v=pred,s=ctf.MCC),bd[,y.class])

        return(list(
          fit=MLG,
          perf.Y=lg.test.Y,
          perf.MCC=lg.test.MCC,
          cutoff.Y=ctf.Y,
          cutoff.MCC=ctf.MCC
        ))
        
      },
      error=function(cond) {
        NA
      }
    )
  },
  mc.cores = detectCores() - 1)
  names(a) = P
  
  # enleve les non convergence qui sont de classe logique (na)
  a[!(sapply(a, is.logical))]
  
  
}





# ---- fonction pour fitter modèle random forests ----
fit.rf <- function(bd, # base de données
                   y, # variable dépendente (caractères)
                   P
)
  
{
  library(randomForest)
  library(doParallel)
  
  

    bd$yf=factor(bd[,y])
      xvars=c(unlist(strsplit(P," + ",fixed=T)),"Variete")
    RF = randomForest(
      x=bd[xvars],
      y=bd$yf,
      data = bd,
      ntree = 501,
      #echantilonnage stratifié
      strata = bd$yf,
      sampsize = c(table(bd$yf)[1],
                   table(bd$yf)[2]),
      importance = T
    )
    RF

}




# ---- fonction pour tester random forests sur  jeu de données de calibration ----
fittest_all.rf <- function(bd0, # base de données
                   y.class, # variable dépendente codée
                   P, # vecteurs des formules des prédicteurs
                   s ,# seuil testé numeric,
                   categ
)
  
{
  library(randomForest)
  library(doParallel)
  
  
  a=mclapply(P, function(x) {
    
    xvars=c(unlist(strsplit(x," \\+ ",fixed=F)),categ)
    bd=na.omit(subset(bd0,select=c(y.class,xvars)))
    bd$yf=factor(bd[,y.class])
    
    RF = randomForest(
      x=bd[xvars],
      y=bd$yf,
      data = bd,
      ntree = 501,
      #echantilonnage stratifié
      strata = bd$yf,
      sampsize = c(table(bd$yf)[1],
                   table(bd$yf)[2]),
      importance = T
    )
    
    pred = predict(RF, newdata = bd, type = "prob")[,2]
    
    # optimisation du cutoff 
    bst=bestcutoff(pred,bd$yf)
    ctf.Y=bst$perf.Y$cutoff
    ctf.MCC=bst$perf.MCC$cutoff
    
    test.Y=perf(rt(v=pred,s=ctf.Y),bd$yf)
    test.MCC=perf(rt(v=pred,s=ctf.MCC),bd$yf)

    return(list(fit=RF,
                ctf.Y=ctf.Y,
                ctf.MCC=ctf.MCC,
                perf.Y=test.Y,
                perf.MCC=test.MCC))
    
    
  },
  mc.cores = detectCores() - 1)
  names(a) = P
  a
}




# ---- fonction pour trouver le meilleur cutoff modeles logistic ----
bestcutoff <- function(pred_prob,y){
  # fonction permettant de trouver le meilleur cutoff pour coder les proba en (0,1)
  # à partir de proba candidate =proba prédites
  # pred_prob = les probas predites
  # y= la classe d'appartenance de la valeur observée (0,1)
  
  ctf=sort(pred_prob)
  
  w=do.call('rbind',
            mclapply(ctf,function(z){
              #predclass=rep(NA,times=length(pred_prob))
              #predclass[pred_prob>z]=1
              #predclass[pred_prob<=z]=0
              predclass=rt(pred_prob,z)
              perf(predclass,y)
            },mc.cores=detectCores()-1))
  
  # performances optimisation cutoff selon indice de Youden
  perf.Y=w[which.max(w$Youden),]
  perf.Y$cutoff=ctf[which.max(w$Youden)]
  
  # performances optimisation cutoff selon indice MCC
  perf.MCC=w[which.max(w$MCC),]
  perf.MCC$cutoff=ctf[which.max(w$MCC)]
  
  return(list(
              perf.Y=perf.Y,
              perf.MCC=perf.MCC))
  
}
# ---- fonction pour calculer la matrice de confusion ----
perf <- function(pred, ref)
  # fonction compare un vecteur 'pred' (0,1)
  # à un vecteur de référence (0,1) 'ref'
  # pour construire la matrice de confusion
  # renvoie les performances
{
  cm = table(pred, ref)
  TP = length(which(pred == 1 & ref == 1))
  FP = length(which(pred == 1 & ref == 0))
  TN = length(which(pred == 0 & ref == 0))
  FN = length(which(pred == 0 & ref == 1))
  
  Acc = (TP + TN) / (TP + TN + FP + FN)
  Err = 1 - Acc
  
  P1M1 = TP / (TP + FN) #Se
  P0M1 = 1 - P1M1
  
  P0M0 = TN / (TN + FP) #Sp
  P1M0 = 1 - P0M0
  
  M1P1 = TP / (TP + FP)
  M0P1 = 1 - M1P1
  
  M0P0 = TN / (TN + FN)
  M1P0 = 1 - M0P0
  
  Youden=P1M1-P1M0
  MCC=sqrt(M1P1*P1M1*P0M0*M0P0)-sqrt(M0P1*P0M1*P1M0*M1P0) #Matthew correlation coef
  
  # retourne les performances pour le cutoff
  data.frame(TP,
             FP,
             TN,
             FN,
             Acc,
             Err,
             P1M1,
             P0M1,
             P0M0,
             P1M0,
             M1P1,
             M0P0,
             M0P1,
             M1P0,
             Youden,
             MCC)
}



# ---- fonction pour calculer les performances d'un modèle de classification ----




perf.mod <- function(x,
                     # le modèle pour la fonction predict
                     test,
                     # data pour la prédiction
                     s = NA,
                     # le seuil
                     ov,
                     # la variable observée pour faire la comparaison
                     type = c("linear", "logistic", "rf") # type de modele à prédire
                     )
                     
                     {
                       if (type == "linear") {
                         pred = 10 ^ predict(x, newdata = test)
                         pred_class=rt(v=pred,s=s)
                         return(perf(pred_class,test[,ov]))

                       }
                       if (type == "logistic") {
                         # prediction de la proba
                         pred = predict(x, newdata = test, type = "response")
                         # optimisation du cutoff et performances
                         return(bestcutoff(pred, ov))
                       }
                       if (type == "rf") {
                         # prediction de la proba
                         pred = predict(x, newdata = test, type = "prob")[,2]
                         # optimisation du cutoff et performances
                         return(bestcutoff(pred, ov))
                       }
                       if (!(type %in% c("linear", "logistic", "rf")))
                       {
                         warning("type de modèle inconnu")
                       }
                       
                       
                       
                     }
                     

# ---- fonction pour faire la validation croisée modeles lineaires mixtes ----
cv.MEM <- function(bd, # dataset
                   doe, # indice pour les dataset train et test de la validation croisée
                   s,
                   y, # variable grain numérique non codée
                   y.class, # variable grain numérique codée
                   P, # vecteur des combinaisons de prédicteurs
                   folder #dossier de sauvegarde des résultats
                   ) {
                   library(doParallel)
  

  i=1:length(P)
                   
                   lapply(i, function(z) {
                     x=P[z]
                     mem.fit.all = fit.lme(
                       bd = bd,
                       y = paste0("log10(", y, ")"),
                       P = x,
                       rand = "1|Variete"
                     )
                     
                     b = mclapply(doe, function(d) {
                       dat.train = bd[d$train, ]
                       dat.test = subset(bd[d$test, ], Variete %in% dat.train$Variete)
                       n.test=nrow(dat.test)
                       
                       # paramétrage modèles linéaires mixtes ----
                       mem.fit = fit.lme(
                         bd = dat.train,
                         y = paste0("log10(", y, ")"),
                         P = x,
                         rand = "1|Variete"
                       )
                       
                       if (length(mem.fit)==1){
                       # test modèles linéaire mixtes ----
                       mem.test=data.frame(
                       pred = 10 ^ predict(mem.fit[[1]], newdata = dat.test),
                       obs=dat.test[,y]
                       )
                       
                       mem.perf=perf(rt(v=mem.test$pred,s=s),dat.test[,y.class])
                       mem.perf$n.test=n.test

                       
      
                       RE=as.data.frame(random.effects(mem.fit[[1]]))
                       RE$Variete=row.names(RE)
                       
                       
                       return(list(
                         train.FE = fixed.effects(mem.fit[[1]]),
                         train.RE = RE,
                         op.test = mem.test,
                         perf=mem.perf
                       ))}
                     },
                     mc.cores = detectCores() - 1)
                     
                     if (length(b)!=0)
                       
                     { 
                       out= list(
                       
                      predicteurs=x, 
                      
                      num_model=z,
                      
                      fit.all=mem.fit.all,
                     
                     FE = do.call('rbind', lapply(b, function(z) {
                       z$train.FE
                     })),
                     
                    RE= do.call('rbind', lapply(b, function(z) {
                      z$train.RE
                    })),
                    
                    obspred.test=do.call('rbind', lapply(b, function(z) {
                      z$op.test
                    })),
                     
                     perf = do.call('rbind', lapply(b, function(z) {
                       z$perf
                     }))
                       )}
                     
                     saveRDS(out,file=paste0(folder,y.class,
                                             "_MEM_",which(P==x),".rds"))
                     
                   })#,
                   #mc.cores = detectCores() - 1)
                   
                  
}


# ---- fonction pour faire la validation croisée modeles logistiques mixtes ----
cv.logit <- cmpfun(function(bd, # dataset
                   doe, # indice pour les dataset train et test de la validation croisée
                   y.class, # variable grain  codée
                   P, # vecteur des combinaisons de prédicteurs,
                   folder #dossier de sauvegarde des résultats
) {
  library(doParallel)
  library(nlme)
  
  i=1:length(P)
  
    lapply(i, function(z) {
      x=P[z]
  
      lg.fit.all = fit.lg(
        bd = bd,
        y =y.class,
        P = x,
        rand = "1|Variete"
      )
      
      ctf=bestcutoff(predict(lg.fit.all[[1]],type='response'),bd[,y.class])
      
      b = mclapply(doe, function(d) {
      dat.train = bd[d$train, ]
      dat.test = subset(bd[d$test, ], Variete %in% dat.train$Variete)
      n.test=nrow(dat.test)
      
      # paramétrage modèles logistic mixtes ----
      lg.fit = fit.lg(
        bd = dat.train,
        y =y.class,
        P = x,
        rand = "1|Variete"
      )
      
      if (length(lg.fit)==1){
        if (nrow(dat.test)>0) {
      # optimisation du cutoff
      pred = predict(lg.fit[[1]], newdata = dat.train, type = "response")
      
      # optimisation du cutoff 
      bst=bestcutoff(pred,dat.train[,y.class])
      ctf.Y=bst$perf.Y$cutoff
      ctf.MCC=bst$perf.MCC$cutoff
      
      # test modèles linéaire mixtes ----
      lg.test = predict(lg.fit[[1]], newdata = dat.test, type = "response")
      lg.test.Y=perf(rt(v=lg.test,s=ctf.Y),dat.test[,y.class])
      lg.test.Y$n.test=n.test
      lg.test.MCC=perf(rt(v=lg.test,s=ctf.MCC),dat.test[,y.class])
      lg.test.MCC$n.test=n.test
        } else {
        
        lg.test.Y=NA
        lg.test.MCC=NA
        
      }
      
      
      return(list(
        train.FE = fixed.effects(lg.fit[[1]]),
        train.RE = data.frame(Variete=row.names(random.effects(lg.fit[[1]])$Variete),
                              RE=unlist(random.effects(lg.fit[[1]])[[1]],use.names=F)),
        test.Y = lg.test.Y,
        test.MCC = lg.test.MCC
      ))}
      
    },
    mc.cores = detectCores() - 1)
    
    if (length(b)!=0)
      
{   out= list(
      
      predicteurs=x, 
      
      num_model=z,
      
      fit.all=lg.fit.all,
      
      cutoff.all.Y=ctf$perf.Y,
      
      cutoff.all.MCC=ctf$perf.MCC,
      
      FE = do.call('rbind', lapply(b, function(z) {
        z$train.FE
      })),
      
      RE = do.call('rbind', lapply(b, function(z) {
        z$train.RE
      })),

      perf.Y = do.call('rbind', lapply(b, function(z) {
        z$test.Y
      })),
      
      perf.MCC = do.call('rbind', lapply(b, function(z) {
        z$test.MCC
      }))
    )}
    
    saveRDS(out,file=paste0(folder,y.class,
                            "_logit_",which(P==x),".rds"))
  })


}
)

# ---- fonction pour faire la validation croisée modeles logistiques effets fixes ----
cv.logit.f <- cmpfun(function(bd, # dataset
                            doe, # indice pour les dataset train et test de la validation croisée
                            y.class, # variable grain  codée
                            P, # vecteur des combinaisons de prédicteurs,
                            categ,# variable effet catégoriel
                            folder #dossier de sauvegarde des résultats
) {
  library(doParallel)

  

  
  lapply(P, function(x) {
   
    
    lg.fit.all = fit.lg.f(
      bd = bd,
      y =y.class,
      P = x,
      categ = categ
    )
    
    ctf=bestcutoff(predict(lg.fit.all,type='response'),bd[,y.class])
    
    b = mclapply(doe, function(d) {
      dat.train = bd[d$train, ]
      dat.test = subset(bd[d$test, ], Variete %in% dat.train$Variete)
      n.test=nrow(dat.test)
      
      # paramétrage modèles logistic mixtes ----
      lg.fit = fit.lg.f(
        bd = dat.train,
        y =y.class,
        P = x,
        categ = categ
      )
      
 
        if (nrow(dat.test)>0) {
          # optimisation du cutoff
          pred = predict(lg.fit, newdata = dat.train, type = "response")
          
          # optimisation du cutoff 
          bst=bestcutoff(pred,dat.train[,y.class])
          ctf.Y=bst$perf.Y$cutoff
          ctf.MCC=bst$perf.MCC$cutoff
          
          # test modèles  ----
          lg.test = predict(lg.fit, newdata = dat.test, type = "response")
          lg.test.Y=perf(rt(v=lg.test,s=ctf.Y),dat.test[,y.class])
          lg.test.Y$n.test=n.test
          lg.test.MCC=perf(rt(v=lg.test,s=ctf.MCC),dat.test[,y.class])
          lg.test.MCC$n.test=n.test
          
          # coefficients
          pm<-data.frame(summary(lg.fit)$coef)
          pm$V=row.names(pm)
          row.names(pm)<-NULL
        } else {
          
          lg.test.Y=NA
          lg.test.MCC=NA
          
        }
        
        
        return(list(
          
          pm=pm,
          test.Y = lg.test.Y,
          test.MCC = lg.test.MCC
        ))
      
    },
    mc.cores = detectCores() - 1)
    

      
    out= list(
      
      predicteurs=x, 
      
      fit.all=lg.fit.all,
      
      perf_all=ctf,
      
      param=do.call('rbind', lapply(b, function(z) {
        z$pm
      })),


      perf.Y = do.call('rbind', lapply(b, function(z) {
        z$test.Y
      })),
      
      perf.MCC = do.call('rbind', lapply(b, function(z) {
        z$test.MCC
      })),
      
      perf.cv=NULL
    )
    
    out$perf.cv=medci(out,"logit")
    
    saveRDS(out,file=paste0(folder,y.class,
                            "_logit_",which(P==x),".rds"))
    rm(out,b)
  })
  
  
}
)

# ---- fonction pour faire la validation croisée modeles multilinéaire ----
cv.ml <- cmpfun(function(bd, # dataset
                              doe, # indice pour les dataset train et test de la validation croisée
                              y, # variable grain numérique
                              y.class, # variable grain  codée
                              s, #seuil en valeur numérique
                              P, # vecteur des combinaisons de prédicteurs,
                              categ,# variable effet catégoriel
                              folder #dossier de sauvegarde des résultats
) {
  library(doParallel)
  
  
  
  
  lapply(P, function(x) {
    
    
    lm.all = fit.lm(
      bd = bd,
      y=y,
      P = x,
      categ = categ
    )
    
    prednat=10^predict(lm.all)
    ml.perf.all=perf(rt(v=prednat,s=s),bd[,y.class])
    
    
    b = mclapply(doe, function(d) {
      dat.train = bd[d$train, ]
      dat.test = subset(bd[d$test, ], Variete %in% dat.train$Variete)
      n.test=nrow(dat.test)
      
      # paramétrage modèles logistic mixtes ----
      lm.fit = fit.lm(
        bd = dat.train,
        y =y,
        P = x,
        categ = categ
      )
      
      
      if (nrow(dat.test)>0) {

        # test modèles  ----
        lm.test = 10^predict(lm.fit,newdata=dat.test)
        lm.test=perf(rt(v=lm.test,s=s),dat.test[,y.class])
        lm.test$n.test=n.test
        
        # coefficients
        pm<-data.frame(summary(lm.fit)$coef)
        pm$V=row.names(pm)
        row.names(pm)<-NULL
      } else {
        
        lm.test=NA
        pm=NA
        
      }
      
      
      return(list(
        
        pm=pm,
        test = lm.test
      ))
      
    },
    mc.cores = detectCores() - 1)
    
    
    
    out= list(
      
      predicteurs=x, 
      
      lm.all=lm.all,
      
      perf_all=ml.perf.all,
      
      param=do.call('rbind', lapply(b, function(z) {
        z$pm
      })),
      
      
      perf= do.call('rbind', lapply(b, function(z) {
        z$test
      })),
      perf.cv=NULL
    )
    out$perf.cv=medci(out,"ML")
    
    saveRDS(out,file=paste0(folder,y.class,
                            "_ML_",which(P==x),".rds"))
    rm(out,b)
  })
  
  
}
)

# ---- fonction pour faire la validation croisée modeles random forest ----
cv.RF <- cmpfun(function(bd, # dataset
                     doe, # indice pour les dataset train et test de la validation croisée
                     y.class, # variable grain  codée
                     P, # vecteur des combinaisons de prédicteurs,
                     folder #dossier de sauvegarde des résultats
) {
  library(doParallel)
  library(randomForest)
 
  mclapply(P, function(x) {
   
    
    rf.fit.all = fit.rf(
      bd = bd,
      y =y.class,
      P = x
    )
    
    ctf=bestcutoff(predict(rf.fit.all,type="prob")[,2],bd[,y.class])
    
    b = mclapply(doe, function(d) {
      dat.train = bd[d$train, ]
      dat.test = subset(bd[d$test, ], Variete %in% dat.train$Variete)
      n.test=nrow(dat.test)
      
      # paramétrage modèles  ----
      rf.fit = fit.rf(
        bd = dat.train,
        y =y.class,
        P = x
      )
      

        if (nrow(dat.test)>0) {
        # optimisation du cutoff
        pred = predict(rf.fit, newdata = dat.train, type = "prob")[,2]
        
        # optimisation du cutoff 
        bst=bestcutoff(pred,dat.train[,y.class])
        ctf.Y=bst$perf.Y$cutoff
        ctf.MCC=bst$perf.MCC$cutoff
        
        # test modèles random forests ----
        rf.test = predict(rf.fit, newdata = dat.test, type = "prob")[,2]
        rf.test.Y=perf(rt(v=rf.test,s=ctf.Y),dat.test[,y.class])
        rf.test.Y$n.test=n.test
        rf.test.MCC=perf(rt(v=rf.test,s=ctf.MCC),dat.test[,y.class])
        rf.test.MCC$n.test=n.test
        
        } else {
        
          
        rf.test.Y=NA
        rf.test.MCC=NA
        
        }
        
        
        return(list(Gini.V=t(importance(rf.fit)[,4]),
                    test.Y = rf.test.Y,
                    test.MCC = rf.test.MCC
        ))
    },
    mc.cores = detectCores() - 1)
    
      out= list(
      
      predicteurs=x, 
      
      fit.all=rf.fit.all,
      
      perf.all=ctf,
      
      Gini.V=do.call('rbind', lapply(b, function(z) {
        z$Gini.V
      })),

      perf.Y = do.call('rbind', lapply(b, function(z) {
        z$test.Y
      })),
      
      perf.MCC = do.call('rbind', lapply(b, function(z) {
        z$test.MCC
      })),
      
      perf.cv=NULL
    )
    
    out$perf.cv=medci(out,"RF")

    saveRDS(out,file=paste0(folder,y.class,
                            "_RF_",which(P==x),".rds"))
    rm(out,b)
  },
  mc.cores = detectCores() - 1)
  

}
)

# ---- fonction area under curve package bayestR ----
auc<- function (x,
                 y,
                 method = c("trapezoid", "step", "spline"),
                 ...)
{
  if (length(x) != length(y)) {
    stop("length x must equal length y")
  }
  idx <- order(x)
  x <- x[idx]
  y <- y[idx]
  switch(
    match.arg(
      arg = method,
      choices = c("trapezoid", "step",
                  "spline")
    ),
    trapezoid = sum((rowMeans(cbind(
      y[-length(y)],
      y[-1]
    ))) * (x[-1] - x[-length(x)])),
    step = sum(y[-length(y)] *
                 (x[-1] - x[-length(x)])),
    spline = stats::integrate(
      stats::splinefun(x,
                       y, method = "natural"),
      lower = min(x),
      upper = max(x)
    )$value
  )
}

# ---- fonction histogramme croissant des P1M1 avec M1P1 ----

plot.perf=function(a,b){
  
  
  a[,"modele"]<-factor(a[,"modele"])
  a[,"modele"]<-reorder(a[,"modele"],a[,"P1M1"],FUN=function(x) mean(x,na.rm=T),order=T)
  m=aggregate(.~modele,FUN=function(x) mean(x,na.rm=T),data=a)
  na<-aggregate(.~modele,FUN=function(x) length(which(is.nan(x))),data=a,na.action="na.pass")
  m<-m[order(m$P1M1),]
  bp=barplot(t(as.matrix(m[,2:3])),beside=T,col=1:2,
             names.arg = m$modele,
             cex.axis=1,cex=0.6,ylim=c(0,1),
             border=NA,
             main=paste0(b," P1M1 (triés) et M1P1 associés\nChiffres = Nb de NaN"))
  grid()
  text(x=bp[1,],t(as.matrix(m[,2:3]))[1,]+0.05,na$P1M1,cex=0.7,srt=45,col=1)
  text(x=bp[1,],t(as.matrix(m[,2:3]))[1,]+0.15,na$M1P1,cex=0.7,srt=45,col=2)
  
  list(m=m,na=na)
  
}


# ---- fonction pour comparer les performances des modèles ----

comp.perf<-function(pth,ctf=T){
  f=list.files(pth)
  
  if (ctf==F){
    
    perf=do.call('rbind',mclapply(f,function(x){
      a=readRDS(paste0(pth,x))
      P=subset(a$perf,select=c("P1M1","M1P1","P0M0","M0P0"))
      
      P$modele=a$predicteurs
      as.data.frame(P)
      
      
    },
    mc.cores=detectCores()-1)
    
    )
    P=plot.perf(perf,"")
   P
    
  } else {
  Y.perf=do.call('rbind',mclapply(f,function(x){
    a=readRDS(paste0(pth,x))
    Y=subset(a$perf.Y,select=c("P1M1","M1P1","P0M0","M0P0"))
    
    Y$modele=a$predicteurs
    as.data.frame(Y)
    
    
  },
  mc.cores=detectCores()-1)
  )
  
  MCC.perf=do.call('rbind',mclapply(f,function(x){
    a=readRDS(paste0(pth,x))
    MCC=subset(a$perf.MCC,select=c("P1M1","M1P1","P0M0","M0P0"))
    MCC$modele=a$num_model
    as.data.frame(MCC)
    
  },
  mc.cores=detectCores()-1)
  )

  par(mfrow=c(2,1))
  Y=plot.perf(Y.perf,"Youden")
  MCC=plot.perf(MCC.perf,"MCC")
  
  par(mfrow=c(1,1))
  
  list(Y.perf=Y,
       MCC.perf=MCC)
  }
}

comp.perf_old2<-function(pth){
  f=list.files(pth)
  Y.perf=do.call('rbind',mclapply(f,function(x){
    a=readRDS(paste0(pth,x))
    Y=apply(subset(a$perf.Y,select=c("P1M1","M1P1","P0M0","M0P0")),2,FUN=function(z){
      mean(z,na.rm=T)
    })

Y$modele=a$num_model
as.data.frame(Y)
    
  },
  mc.cores=detectCores()-1)
  )
  
  MCC.perf=do.call('rbind',mclapply(f,function(x){
    a=readRDS(paste0(pth,x))
    MCC=apply(subset(a$perf.MCC,select=c("P1M1","M1P1","P0M0","M0P0")),2,FUN=function(z){
      mean(z,na.rm=T)
    })
    
    MCC$modele=a$num_model
    as.data.frame(MCC)
    
  },
  mc.cores=detectCores()-1)
  )
  
  
  
  
  par(mfrow=c(2,2),las=1)
  plot(Y.perf$P1M1,Y.perf$M1P1,pch=19,main="Youden",xlab="P1M1",ylab="M1P1")
  text(Y.perf$P1M1,Y.perf$M1P1,Y.perf$modele,pos=1:4,cex=0.5)
  
  plot(Y.perf$P0M0,Y.perf$M0P0,pch=19,main="Youden",xlab="P0M0",ylab="M0P0")
  text(Y.perf$P0M0,Y.perf$M0P0,Y.perf$modele,pos=1:4,cex=0.5)  
  
  plot(MCC.perf$P1M1,MCC.perf$M1P1,pch=19)
  text(MCC.perf$P1M1,MCC.perf$M1P1,MCC.perf$modele,pos=1:4,cex=0.5)
  
  plot(MCC.perf$P0M0,MCC.perf$M0P0,pch=19)
  text(MCC.perf$P0M0,MCC.perf$M0P0,MCC.perf$modele,pos=1:4,cex=0.5)  

  par(mfrow=c(1,1))
}
comp.perf_old<-function(pth){
  f=list.files(pth)
  w=do.call('rbind',mclapply(f,function(x){
    a=readRDS(paste0(pth,x))
    Y=a$perf.Y$Youden
    f.Y=ecdf(Y)
    auc.Y=auc(na.omit(Y),f.Y(na.omit(Y)))
    na.Y=1-length(na.omit(Y))/length(Y)
    
    MCC=a$perf.MCC$MCC
    f.MCC=ecdf(MCC)
    auc.MCC=auc(na.omit(MCC),f.MCC(na.omit(MCC)))
    na.MCC=1-length(na.omit(MCC))/length(MCC)
    
    
    
    
    data.frame(auc.Y=auc.Y,
                 na.Y=na.Y,
                 auc.MCC=auc.MCC,
                 na.MCC=na.MCC,
                 modele=a$num_model)
    
  },
  mc.cores=detectCores()-1)
  )


  par(mfrow=c(2,1),las=1)
plot(w$na.Y,w$auc.Y,pch=19,main="Youden",
     xlab="Freq des NaN",ylab="AUC ecdf(Youden)")
text(w$na.Y,w$auc.Y,w$modele,pos=1:4,cex=0.5)
  
plot(w$na.MCC,w$auc.MCC,pch=19,main="MCC",
     xlab="Freq des NaN",ylab="AUC ecdf(MCC)")
text(w$na.MCC,w$auc.MCC,w$modele,pos=1:4,cex=0.5)
  par(mfrow=c(1,1))
}

# ---- fonction calcul intervalle confiance médiane Y et MCC sur les sorties de la validation croisée ----
medci=function(a,approach,alpha=0.05){
  library(boot)
  # fonction pour calculer la médiane en enlevant les NA
  med=function(x,y) median(x[y],na.rm=T)
  
  
  # pour les modèle MEM il n'y a pas eu optimisation de cutoff
  # il n'y a qu'un seul objet pour les performances du modèle 
  # pour logit et RF il y a eu optimisation du cutoff
  # l'indice de Youden est pris dans l'objet perf.Y correspondant
  # à une optiisation du cutoff pour maximiser Youden
  # l'indice MCC est pris dans l'objet perf.M correspondant
  # à une optiisation du cutoff pour maximiser MCC
  
  if (approach=="ML") {
    dY = a$perf
    dMCC = a$perf
    
    #on rajoute le calcul des medianes et mad des variable de performances
    perf.Y.med=apply(a$perf,2,function(y) median(y, na.rm=T))
    perf.Y.mad=apply(a$perf,2,function(y) mad(y, na.rm=T))
    
    perf.MCC.med=apply(a$perf,2,function(y) median(y, na.rm=T)) # copie de perf.Y
    perf.MCC.mad=apply(a$perf,2,function(y) mad(y, na.rm=T))
    
    
  } else {
    perf.Y.med=apply(a$perf.Y,2,function(y) median(y, na.rm=T))
    perf.Y.mad=apply(a$perf.Y,2,function(y) mad(y, na.rm=T))
    
    perf.MCC.med=apply(a$perf.MCC,2,function(y) median(y, na.rm=T))
    perf.MCC.mad=apply(a$perf.MCC,2,function(y) mad(y, na.rm=T))
    
    dY = a$perf.Y
    dMCC = a$perf.MCC
  }
  
  names(perf.Y.med)=paste0(names(perf.Y.med),".med")
  names(perf.Y.mad)=paste0(names(perf.Y.mad),".mad")
  names(perf.MCC.med)=paste0(names(perf.MCC.med),".med")
  names(perf.MCC.mad)=paste0(names(perf.MCC.mad),".mad")
  
  # boostrap pour calcul de la médiane de Youden 1000 tirages
  Y=boot(dY$Youden,statistic=med,R=1e3,
         parallel="multicore",ncpus=detectCores()-1)
  
  # calcul de l'interval de confiance basic
  # (on ne fait pas l'hypothèse d'une distribution normale)
  Yci=boot.ci(Y,type="basic",conf=1-alpha)
  
  MCC=boot(dMCC$MCC,statistic=med,R=1e3,
           parallel="multicore",ncpus=detectCores()-1)
  MCCci=boot.ci(MCC,type="basic",conf=1-alpha)
  
  # dataframe des sorties 
  return(list(
    Y_IC=cbind(data.frame(
    model=a$predicteurs,
    approache=approach,  
    med.Y=Y$t0, # mediane estimée par bootstrap
    Ylwr.ci=Yci$basic[4], # borne sup de l'intervalle de confiance
    Yupr.ci=Yci$basic[5]), # borne inf de l'intervalle de confiance
    t(perf.Y.med),
    t(perf.Y.mad)
    ),
    
    MCC_IC=cbind(data.frame(
      model=a$predicteurs,
      approache=approach,  
    med.MCC=MCC$t0,
    MCClwr.ci=MCCci$basic[4],
    MCCupr.ci=MCCci$basic[5]),
    t(perf.MCC.med),
    t(perf.MCC.mad)
)
  ))

}
# ---- fonction compilant les performances medianes issues de la validation croisées pour tous les modeles et les approches-----
compil.perf<-function(pth){
  library(boot)
  library(doParallel)
  
  med=function(x,y) median(x[y],na.rm=T)
  f=list.files(c(paste0(pth,"ML"),paste0(pth,"logit"),paste0(pth,"RF")),full.names = T)
  
  perfCV.Y<-do.call('rbind',lapply(f,function(x) {
    readRDS(x)$perf.cv$Y_IC
          
  }))
  
  

  perfCV.MCC<-do.call('rbind',lapply(f,function(x) {
    readRDS(x)$perf.cv$MCC_IC
  }))
 
  
  
  return(list(perfCV.Y=perfCV.Y,
              perfCV.MCC=perfCV.MCC
  ))
  

  }


# ---- fonction pour la simulation shiny ----
simule_all_best_models<-function(In,mod,val.centrage){
  
  #----preparation donnee entree----
  
  In$Heau=10^-In$pHeau
  
  #---- texture 3 fractions ----
  
  if (is.na(In$Lim)==F) {
    
    a=subset(In, select=intersect(names(val.centrage),names(In)))
    a=a[which(is.na(a)==F)]
    inpar0=rbind(log10(a),val.centrage[names(val.centrage)%in%names(a)])
    inpar=inpar0[1,]-inpar0[2,]
    
    names(inpar)=paste0(names(inpar),".lc")
    inpar=cbind(inpar,a)
    inpar$Variete=In$Variete
    inpar$seuil=In$seuil
    rm(inpar0,a)
    
    modsim=mod$bestLim
  }
  
  #---- texture 5 fractions ----
  if (is.na(In$Limf)==F) {
    
    a=subset(In, select=intersect(names(val.centrage),names(In)))
    a=a[which(is.na(a)==F)]
    inpar0=rbind(log10(a),val.centrage[names(val.centrage)%in%names(a)])
    inpar=inpar0[1,]-inpar0[2,]
    
    names(inpar)=paste0(names(inpar),".lc")
    inpar=cbind(inpar,a)
    inpar$Variete=In$Variete
    inpar$seuil=In$seuil
    rm(inpar0,a)
    
    
    modsim=mod$bestLimf
  }
  #---- pas de données texture----
  
  if (is.na(In$Arg)==T) {
    a=subset(In, select=intersect(names(val.centrage),names(In)))
    a=a[which(is.na(a)==F)]
    inpar0=rbind(log10(a),val.centrage[names(val.centrage)%in%names(a)])
    inpar=inpar0[1,]-inpar0[2,]
    
    names(inpar)=paste0(names(inpar),".lc")
    inpar=cbind(inpar,a)
    inpar$Variete=In$Variete
    inpar$seuil=In$seuil
    rm(inpar0,a)
    
    modsim=mod$bestnotext
  }
  
  
  
  #---- simulations ----
  
  simY=mclapply(modsim,function(z){
    #par Y, MCC
    
    
    
    if (class(z$fit.all)[1] == "randomForest") {
      p = predict(z$fit.all, newdata = inpar, type = 'prob')[2]
      
      ctf = z$perf.all$perf.Y$cutoff
      if (p > ctf) {
        predY = 1
      } else {
        predY = 0
      }
      if (p <= ctf) {
        pY.pond = p * 0.5 / ctf
      } else {
        pY.pond = (p - ctf) * 0.5 / (1 - ctf) + 0.5
      }
    }
    
    if (class(z$fit.all)[1]  == "glm") {
      p = predict(z$fit.all, newdata = inpar, type = 'response')
      
      ctf = z$perf_all$perf.Y$cutoff
      if (p > ctf) {
        predY = 1
      } else {
        predY = 0
      }
      if (p <= ctf) {
        pY.pond = p * 0.5 / ctf
      } else {
        pY.pond = (p - ctf) * 0.5 / (1 - ctf) + 0.5
      }
    }
    
    if (class(z$fit.all)[1] == "lm") {
      p = 10 ^ predict(z$fit.all, newdata = inpar)
      ctf = inpar$seuil
      if (p > ctf) {
        predY = 1
      } else {
        predY = 0
      }
      pY.pond=NA
    }
    
    cbind(data.frame(seuil=In$seuil,
                     predY=predY,
                     pY.pond=pY.pond,
                     predicteurs=z$predicteurs,
                     approach=class(z$fit.all)[1]),
          (z$perf.cv$Y_IC[c( "P1M1.med", "P0M1.med", "P0M0.med", "P1M0.med", "M1P1.med", "M0P0.med", "M0P1.med", "M1P0.med")])
    )
    
    
  },mc.cores = detectCores()-1)
  simY=do.call('rbind',simY)
  
  synthY=data.frame(seuil=unique(simY$seuil),
                    nb.models=nrow(simY),
                    risk.nc=median(simY$pY.pond,na.rm=T),
                    risk.nc.mad=mad(simY$pY.pond,na.rm=T),
                    predY.med=median(simY$predY),
                    predY.mad=mad(simY$predY),
                    
                    t(apply(simY[,c("pY.pond","P1M1.med", "P0M1.med", "P0M0.med", "P1M0.med", "M1P1.med", "M0P0.med", "M0P1.med", "M1P0.med")],2,
                            FUN=function(b) median(b,na.rm=T))))
  
  
  
  
  list(synthY=synthY,
       simY =simY )
}